#include "draw.h"
#include "gpu.h"
#include "stdatomic.h"
#include "vulkan/vulkan_core.h"

void record_ui_draw(VkCommandBuffer command_buffer, UIContext* ui_context, double time, uint32_t frame) {
  UIPushConstant push = {
    .time = (float)time,
    .layer = 0,
  };

  vkCmdBindPipeline(command_buffer, VK_PIPELINE_BIND_POINT_GRAPHICS, ui_context->pipeline.pipeline);
  vkCmdBindDescriptorSets(command_buffer, VK_PIPELINE_BIND_POINT_GRAPHICS, ui_context->pipeline.layout, 0, 1, &ui_context->font_samplers, 0, NULL);
  vkCmdBindDescriptorSets(command_buffer, VK_PIPELINE_BIND_POINT_GRAPHICS, ui_context->pipeline.layout, 1, 1, &ui_context->font_textures, 0, NULL);
  vkCmdBindDescriptorSets(command_buffer, VK_PIPELINE_BIND_POINT_GRAPHICS, ui_context->pipeline.layout, 2, 1, &ui_context->samplers, 0, NULL);
  vkCmdBindDescriptorSets(command_buffer, VK_PIPELINE_BIND_POINT_GRAPHICS, ui_context->pipeline.layout, 3, 1, &ui_context->textures, 0, NULL);
  for(uint32_t i = 0; i < ui_context->max_containers; i++) {
    uint32_t layer_count = atomic_load(&ui_context->containers[i].layer_count);
    __sync_synchronize();
    for(uint32_t j = 0; j < layer_count; j++) {
      push.layer = ui_context->containers[i].layers[j].address[frame];
      vkCmdPushConstants(command_buffer, ui_context->pipeline.layout, VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_FRAGMENT_BIT, 0, 16, &push);
      vkCmdDrawIndirect(command_buffer, ui_context->containers[i].layers[j].layer[frame], offsetof(GPULayer, draw), 1, 0);
    }
  }
}

void record_ui_compute(VkCommandBuffer command_buffer, UIContext* ui, uint32_t frame) {
  UIPushConstant push = {
    .time = 0.0,
    .layer = 0,
  };

  vkCmdBindPipeline(command_buffer, VK_PIPELINE_BIND_POINT_COMPUTE, ui->string_pipeline.pipeline);
  for(uint32_t i = 0; i < ui->max_containers; i++) {
    if(ui->containers[i].id != 0x00000000) {
      for(uint32_t j = 0; j < ui->containers[i].layer_count; j++) {
        push.layer = ui->containers[i].layers[j].address[frame];
        command_copy_buffer(command_buffer, ui->containers[i].layers[j].layer[frame], ui->containers[i].layers[j].layer[frame], offsetof(GPULayer, num_drawables), offsetof(GPULayer, draw) + offsetof(DrawCommand, instance_count), sizeof(uint32_t));
        vkCmdPushConstants(command_buffer, ui->string_pipeline.layout, VK_SHADER_STAGE_COMPUTE_BIT, 0, 16, &push);
        vkCmdDispatchIndirect(command_buffer, ui->containers[i].layers[j].layer[frame], offsetof(GPULayer, dispatch_strings));
      }
    }
  }
}

VkResult draw_frame(
    RenderContext* context,
    UIContext* ui,
    double time) {
  VkResult result;

  VkFence fences[] = {context->frame[context->current_frame].ready};
  VK_RESULT(vkWaitForFences(context->device, 1, fences, VK_TRUE, UINT64_MAX));
  VK_RESULT(vkResetFences(context->device, 1, fences));

  VkCommandBufferBeginInfo begin_info = {
    .sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO,
  }; 

  if(context->frame[context->current_frame].transfer_count > 0) {
    VkFence fences[] = {context->frame[context->current_frame].transfer_ready[0], context->frame[context->current_frame].transfer_ready[1]};
    VK_RESULT(vkWaitForFences(context->device, 2, fences, VK_TRUE, UINT64_MAX));
    VK_RESULT(vkResetFences(context->device, 2, fences));

    for(uint32_t i = 0; i < MAX_FRAMES_IN_FLIGHT; i++) {
      VkCommandBuffer transfer_commands = context->frame[context->current_frame].transfer_commands[i];
      VK_RESULT(vkResetCommandBuffer(transfer_commands, 0));
      VK_RESULT(vkBeginCommandBuffer(transfer_commands, &begin_info));

      VkDeviceSize src_offset = 0;
      for(uint32_t j = 0; j < context->frame[context->current_frame].transfer_count; j++) {
        command_copy_buffer(
            transfer_commands,
            context->frame[context->current_frame].transfer_buffer,
            context->frame[context->current_frame].transfer_infos[j].buffers[i],
            src_offset,
            context->frame[context->current_frame].transfer_infos[j].dst_offset,
            context->frame[context->current_frame].transfer_infos[j].size);
        src_offset += context->frame[context->current_frame].transfer_infos[j].size;
      }
      record_ui_compute(transfer_commands, ui, i);
      VK_RESULT(vkEndCommandBuffer(transfer_commands));
      VkPipelineStageFlags wait_stages[] = {VK_PIPELINE_STAGE_ALL_COMMANDS_BIT, VK_PIPELINE_STAGE_ALL_COMMANDS_BIT};
      context->frame[i].transfer_index += 1;
      VkSemaphore transfer_signals[] = {context->frame[i].transfer};
      uint64_t transfer_signal_values[] = {context->frame[i].transfer_index};
      VkSemaphore transfer_waits[] = {context->frame[i].transfer, context->frame[i].frame};
      uint64_t transfer_wait_values[] = {context->frame[i].transfer_index-1, context->frame[i].frame_index};
      VkTimelineSemaphoreSubmitInfo timeline_info = {
        .sType = VK_STRUCTURE_TYPE_TIMELINE_SEMAPHORE_SUBMIT_INFO,
        .signalSemaphoreValueCount = sizeof(transfer_signal_values)/sizeof(uint64_t),
        .pSignalSemaphoreValues = transfer_signal_values,
        .waitSemaphoreValueCount = sizeof(transfer_wait_values)/sizeof(uint64_t),
        .pWaitSemaphoreValues = transfer_wait_values,
      };
      VkSubmitInfo submit_info = {
        .sType = VK_STRUCTURE_TYPE_SUBMIT_INFO,
        .commandBufferCount = 1,
        .pCommandBuffers = &transfer_commands,
        .pSignalSemaphores = transfer_signals,
        .signalSemaphoreCount = sizeof(transfer_signals)/sizeof(VkSemaphore),
        .pWaitSemaphores = transfer_waits,
        .pWaitDstStageMask = wait_stages,
        .waitSemaphoreCount = sizeof(transfer_waits)/sizeof(VkSemaphore),
        .pNext = &timeline_info,
      };
      VK_RESULT(vkQueueSubmit(context->transfer_queue.handle, 1, &submit_info, context->frame[context->current_frame].transfer_ready[i]));
    }
    context->frame[context->current_frame].transfer_count = 0;
  }

  uint32_t image_index;
  result = vkAcquireNextImageKHR(context->device, context->swapchain, UINT64_MAX, context->frame[context->current_frame].image, VK_NULL_HANDLE, &image_index);
  if(result != VK_SUCCESS) {
    return result;
  }

  VkCommandBuffer command_buffer = context->swapchain_command_buffers[image_index];
  VK_RESULT(vkResetCommandBuffer(command_buffer, 0));
  VK_RESULT(vkBeginCommandBuffer(command_buffer, &begin_info));

  VkViewport viewport = {
    .width = context->swapchain_extent.width,
    .height = context->swapchain_extent.height,
    .maxDepth = 1.0f,
    .minDepth = 0.0f,
  };
  vkCmdSetViewport(command_buffer, 0, 1, &viewport);

  VkRect2D scissor = {
    .extent = context->swapchain_extent,
  };
  vkCmdSetScissor(command_buffer, 0, 1, &scissor);

  VkClearValue clear_values[2] = {{.color={{0.0f, 0.0f, 0.0f, 0.0f}}}, {.depthStencil={1.0f, 0.0f}}};

  VkRenderPassBeginInfo render_pass_begin = {
    .sType = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO,
    .renderPass = context->render_pass,
    .framebuffer = context->swapchain_framebuffers[image_index],
    .renderArea.offset = {0, 0},
    .renderArea.extent = context->swapchain_extent,
    .clearValueCount = 2,
    .pClearValues = clear_values,
  }; 

  vkCmdBeginRenderPass(command_buffer, &render_pass_begin, VK_SUBPASS_CONTENTS_INLINE);
  vkCmdNextSubpass(command_buffer, VK_SUBPASS_CONTENTS_INLINE);
  record_ui_draw(command_buffer, ui, time, context->current_frame);
  vkCmdEndRenderPass(command_buffer);

  VK_RESULT(vkEndCommandBuffer(command_buffer));

  VkPipelineStageFlags wait_stages[] = {VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT, VK_PIPELINE_STAGE_DRAW_INDIRECT_BIT};
  VkSemaphore wait_semaphores[] = {context->frame[context->current_frame].image, context->frame[context->current_frame].transfer};
  VkSemaphore signal_semaphores[] = {context->frame[context->current_frame].render, context->frame[context->current_frame].frame};
  context->frame[context->current_frame].frame_index += 1;
  uint64_t wait_values[] = {0, context->frame[context->current_frame].transfer_index};
  uint64_t signal_values[] = {0, context->frame[context->current_frame].frame_index};
  VkTimelineSemaphoreSubmitInfo timeline_info = {
    .sType = VK_STRUCTURE_TYPE_TIMELINE_SEMAPHORE_SUBMIT_INFO,
    .signalSemaphoreValueCount = sizeof(signal_values)/sizeof(uint64_t),
    .pSignalSemaphoreValues = signal_values,
    .waitSemaphoreValueCount = sizeof(wait_values)/sizeof(uint64_t),
    .pWaitSemaphoreValues = wait_values,
  };
  VkSubmitInfo submit_info = {
    .sType = VK_STRUCTURE_TYPE_SUBMIT_INFO,
    .waitSemaphoreCount = sizeof(wait_semaphores)/sizeof(VkSemaphore),
    .pWaitSemaphores = wait_semaphores,
    .pWaitDstStageMask = wait_stages,
    .commandBufferCount = 1,
    .pCommandBuffers = &context->swapchain_command_buffers[image_index],
    .signalSemaphoreCount = sizeof(signal_semaphores)/sizeof(VkSemaphore),
    .pSignalSemaphores = signal_semaphores,
    .pNext = &timeline_info,
  };

  result = vkQueueSubmit(context->graphics_queue.handle, 1, &submit_info, context->frame[context->current_frame].ready);
  if(result != VK_SUCCESS) {
    return result;
  }

  VkPresentInfoKHR present_info = {
    .sType = VK_STRUCTURE_TYPE_PRESENT_INFO_KHR,
    .waitSemaphoreCount = 1,
    .pWaitSemaphores = &context->frame[context->current_frame].render,
    .swapchainCount = 1,
    .pSwapchains = &context->swapchain,
    .pImageIndices = &image_index,
    .pResults = 0,
  };

  result = vkQueuePresentKHR(context->present_queue.handle, &present_info);
  if(result != VK_SUCCESS) {
    return result;
  }

  context->current_frame = (context->current_frame + 1) % MAX_FRAMES_IN_FLIGHT;

  return VK_SUCCESS;
}
